@@ -5,7 +5,7 @@
 
 use self::{overlay::*, tracing::*, voxel::*};
 use bevy::{
-    core_pipeline,
+    core_pipeline::{self, AlphaMask3d, Opaque3d, Transparent3d},
     prelude::*,
     reflect::TypeUuid,
     render::{
@@ -19,11 +19,13 @@ use bevy::{
 };
 
 mod overlay;
+mod storage_vec;
 mod tracing;
 mod voxel;
 
 pub const VOXEL_SIZE: usize = 256;
 pub const VOXEL_ANISOTROPIC_MIPMAP_LEVEL_COUNT: usize = 8;
+pub const VOXEL_COUNT: usize = 16777216;
 
 pub const VOXEL_SHADER_HANDLE: HandleUntyped =
     HandleUntyped::weak_from_u64(Shader::TYPE_UUID, 14750151725749984740);
@@ -42,6 +44,7 @@ pub mod draw_3d_graph {
     }
 }
 
+pub use storage_vec::StorageVec;
 pub use tracing::TracingMaterialPlugin;
 pub use voxel::VoxelMaterialPlugin;
 
@@ -184,6 +187,7 @@ impl Plugin for VoxelConeTracingPlugin {
 pub enum VoxelConeTracingSystems {
     ExtractVolumes,
     ExtractViews,
+    ExtractReceiverFilter,
     PrepareVolumes,
     QueueVolumeViewBindGroups,
     QueueVoxelBindGroups,
@@ -193,6 +197,14 @@ pub enum VoxelConeTracingSystems {
     QueueTracingBindGroups,
 }
 
+/// Marker component for meshes not casting GI.
+#[derive(Component)]
+pub struct NotGiCaster;
+
+/// Marker component for meshes not receiving GI.
+#[derive(Component)]
+pub struct NotGiReceiver;
+
 /// A component attached to [`Camera`] to indicate the volume of voxelization.
 #[derive(Component, Clone)]
 pub struct Volume {
@@ -232,6 +244,11 @@ pub struct VolumeUniformOffset {
     pub offset: u32,
 }
 
+#[derive(Component, Clone)]
+pub struct VoxelBufferOffset {
+    pub offset: u32,
+}
+
 #[derive(Component)]
 pub struct VolumeColorAttachment {
     pub texture: CachedTexture,
@@ -251,9 +268,20 @@ pub struct GpuVolume {
     max: Vec3,
 }
 
+#[derive(Clone, AsStd140)]
+pub struct GpuVoxel {
+    color: [u32; 2],
+}
+
+#[derive(AsStd140)]
+pub struct GpuVoxelBuffer {
+    data: [GpuVoxel; VOXEL_COUNT],
+}
+
 #[derive(Default)]
 pub struct VolumeMeta {
     volume_uniforms: DynamicUniformVec<GpuVolume>,
+    voxel_buffers: StorageVec<GpuVoxelBuffer>,
 }
 
 pub fn add_volume_overlay(
@@ -323,6 +351,7 @@ pub fn prepare_volumes(
     mut volume_meta: ResMut<VolumeMeta>,
 ) {
     volume_meta.volume_uniforms.clear();
+    volume_meta.voxel_buffers.clear();
 
     for (entity, volume, overlay) in volumes.iter_mut() {
         let volume_uniform_offset = VolumeUniformOffset {
@@ -332,6 +361,10 @@ pub fn prepare_volumes(
             }),
         };
 
+        let voxel_buffer_offset = VoxelBufferOffset {
+            offset: volume_meta.voxel_buffers.push(&render_device),
+        };
+
         let voxel_texture = texture_cache.get(
             &render_device,
             TextureDescriptor {
@@ -392,7 +425,7 @@ pub fn prepare_volumes(
                         depth_or_array_layers: 1,
                     },
                     mip_level_count: 1,
-                    sample_count: 4,
+                    sample_count: 1,
                     dimension: TextureDimension::D2,
                     format: TextureFormat::Bgra8UnormSrgb,
                     usage: TextureUsages::RENDER_ATTACHMENT,
@@ -401,6 +434,7 @@ pub fn prepare_volumes(
 
             commands.entity(view).insert_bundle((
                 volume_uniform_offset.clone(),
+                voxel_buffer_offset.clone(),
                 VolumeColorAttachment {
                     texture: color_texture,
                 },
@@ -423,13 +457,16 @@ pub fn prepare_volumes(
 
         commands.entity(entity).insert_bundle((
             volume_uniform_offset.clone(),
+            voxel_buffer_offset.clone(),
             VolumeBindings {
                 overlay_depth_texture,
                 voxel_texture,
                 anisotropic_textures,
                 texture_sampler,
             },
-            RenderPhase::<Tracing>::default(),
+            RenderPhase::<Tracing<Opaque3d>>::default(),
+            RenderPhase::<Tracing<AlphaMask3d>>::default(),
+            RenderPhase::<Tracing<Transparent3d>>::default(),
         ));
     }
 
